from math import isclose
import unittest
from pathlib import Path

from accelforge.frontend import Spec
from accelforge.frontend.mapping import Mapping
from accelforge.frontend.workload import Workload

from accelforge.mapper.FFM._make_pmappings.pmapper_job import Job
from accelforge.model._looptree.accesses import (
    isl_buffer_accesses_from_buffet_actions,
    Accesses,
)
from accelforge.model._looptree.energy import gather_actions
from accelforge.model._looptree.latency import get_latency
from accelforge.model._looptree.types import Buffet
from accelforge.model._looptree.reuse.symbolic import (
    BuffetStats,
    analyze_reuse_and_add_reservations_to_mapping,
    Compute,
)


PARENT_DIR = Path(__file__).parent


def make_job(mapping: Mapping, workload: Workload) -> Job:
    return Job(
        spec=None,
        mapping=mapping,
        workload=workload,
    )


class TestSymbolicModel(unittest.TestCase):
    def test_q_mapping(self):
        mapping = Mapping.from_yaml(Path(__file__).parent / "Q_mapping.yaml")
        workload = Workload.from_yaml(Path(__file__).parent / "mha.yaml")

        result = analyze_reuse_and_add_reservations_to_mapping(
            make_job(mapping, workload)
        )

        self.assertAlmostEqual(
            result.compute_stats[Compute("Q", "MAC")].total_ops, 64.0
        )
        self.assertAlmostEqual(
            result.compute_stats[Compute("Q", "MAC")].max_per_unit_ops, 16.0
        )

    def test_conv_mapping(self):
        mapping = Mapping.from_yaml(Path(__file__).parent / "conv.mapping.yaml")
        workload = Workload.from_yaml(Path(__file__).parent / "conv.workload.yaml")

        result = analyze_reuse_and_add_reservations_to_mapping(
            make_job(mapping, workload)
        )

        self.assertAlmostEqual(
            result.compute_stats[Compute("conv", "MAC")].total_ops, 120.0
        )
        self.assertAlmostEqual(
            result.compute_stats[Compute("conv", "MAC")].max_per_unit_ops, 10.0
        )

    def test_matmul_mapping(self):
        mapping = Mapping.from_yaml(Path(__file__).parent / "matmul.mapping.yaml")
        workload = Workload.from_yaml(Path(__file__).parent / "matmul.workload.yaml")

        result = analyze_reuse_and_add_reservations_to_mapping(
            make_job(mapping, workload)
        )

        REF_OCCUPANCY = {"W0": 1, "T0": 128, "T1": 128 * 128}
        for tensor, ref_occupancy in REF_OCCUPANCY.items():
            self.assertAlmostEqual(
                result.buffet_stats[
                    Buffet(tensor, "Matmul1", "LocalBuffer")
                ].max_occupancy,
                ref_occupancy,
            )

    def test_matmul_spatial(self):
        mapping = Mapping.from_yaml(PARENT_DIR / "matmul_spatial.mapping.yaml")
        workload = Workload.from_yaml(PARENT_DIR / "matmul.workload.yaml")

        result = analyze_reuse_and_add_reservations_to_mapping(
            make_job(mapping, workload)
        )
        self.assertAlmostEqual(
            result.fanout,
            {
                ("LocalBuffer", "Matmul1"): {0: 128.0, 1: 4.0},
                ("MainMemory", "Matmul1"): {},
            },
        )

    def test_copy_mapping(self):
        mapping = Mapping.from_yaml(PARENT_DIR / "copy.mapping.yaml")
        workload = Workload.from_yaml(PARENT_DIR / "copy.workload.yaml")

        result = analyze_reuse_and_add_reservations_to_mapping(
            make_job(mapping, workload)
        )

        self.assertAlmostEqual(
            result.compute_stats[Compute("copy", "MAC")].total_ops, 0
        )
        self.assertAlmostEqual(
            result.compute_stats[Compute("copy", "MAC")].max_per_unit_ops, 0
        )

        for tensor in ["O1", "O2", "O3", "O4"]:
            for memory in [
                "MainMemory",
                "GlobalBuffer",
                "LocalBuffer",
                "Register",
                "MAC",
            ]:
                buffet = Buffet(level=memory, tensor=tensor, einsum="copy")
                if buffet not in result.buffet_stats:
                    continue
                stats = result.buffet_stats[buffet]
                self.assertAlmostEqual(stats.net_total_read_actions(), 0)
                self.assertAlmostEqual(stats.net_max_per_unit_read_actions(), 0)
                self.assertAlmostEqual(stats.net_total_write_actions(), 0)
                self.assertAlmostEqual(stats.net_max_per_unit_write_actions(), 0)
                self.assertAlmostEqual(stats.max_occupancy, 0)

        buffet = Buffet(level="GlobalBuffer", tensor="I", einsum="copy")
        stats = result.buffet_stats.get(buffet, BuffetStats())
        self.assertAlmostEqual(stats.net_total_read_actions(), 0)
        self.assertAlmostEqual(stats.net_max_per_unit_read_actions(), 0)
        self.assertAlmostEqual(stats.net_total_write_actions(), 0)
        self.assertAlmostEqual(stats.net_max_per_unit_write_actions(), 0)
        self.assertAlmostEqual(stats.max_occupancy, 0)

        stats = result.buffet_stats[
            Buffet(level="LocalBuffer", tensor="I", einsum="copy")
        ]
        self.assertAlmostEqual(stats.net_total_read_actions(), 16)
        self.assertAlmostEqual(stats.net_max_per_unit_read_actions(), 16)
        self.assertAlmostEqual(stats.net_total_write_actions(), 0)
        self.assertAlmostEqual(stats.net_max_per_unit_write_actions(), 0)
        self.assertAlmostEqual(stats.max_occupancy, 2)

        stats = result.buffet_stats[Buffet(level="Register", tensor="I", einsum="copy")]
        self.assertAlmostEqual(stats.net_total_read_actions(), 0)
        self.assertAlmostEqual(stats.net_max_per_unit_read_actions(), 0)
        self.assertAlmostEqual(stats.net_total_write_actions(), 8)
        self.assertAlmostEqual(stats.net_max_per_unit_write_actions(), 8)
        self.assertAlmostEqual(stats.max_occupancy, 1)

        stats = result.buffet_stats[
            Buffet(level="MainMemory", tensor="I", einsum="copy")
        ]
        self.assertAlmostEqual(stats.net_total_read_actions(), 8)
        self.assertAlmostEqual(stats.net_max_per_unit_read_actions(), 8)
        self.assertAlmostEqual(stats.net_total_write_actions(), 8)
        self.assertAlmostEqual(stats.net_max_per_unit_write_actions(), 8)
        self.assertAlmostEqual(stats.max_occupancy, 8)

        stats = result.buffet_stats[Buffet(level="Disk", tensor="I", einsum="copy")]
        self.assertAlmostEqual(stats.net_total_read_actions(), 0)
        self.assertAlmostEqual(stats.net_max_per_unit_read_actions(), 0)
        self.assertAlmostEqual(stats.net_total_write_actions(), 8)
        self.assertAlmostEqual(stats.net_max_per_unit_write_actions(), 8)
        self.assertAlmostEqual(stats.max_occupancy, 8)


class TestSymbolicAccesses(unittest.TestCase):
    def test_q_mapping(self):
        mapping = Mapping.from_yaml(Path(__file__).parent / "Q_mapping.yaml")
        workload = Workload.from_yaml(Path(__file__).parent / "mha.yaml")

        result = analyze_reuse_and_add_reservations_to_mapping(
            make_job(mapping, workload)
        )

        # main_memory_I_accesses = accesses.get_accesses('MainMemory', 'I', 'Q')
        stats = result.buffet_stats[Buffet(level="MainMemory", tensor="I", einsum="Q")]
        self.assertAlmostEqual(stats.net_total_read_actions(), 64.0)
        self.assertAlmostEqual(stats.net_max_per_unit_read_actions(), 64.0)
        self.assertAlmostEqual(stats.net_total_write_actions(), 0.0)
        self.assertAlmostEqual(stats.net_max_per_unit_write_actions(), 0.0)

        stats = result.buffet_stats[Buffet(level="LocalBuffer", tensor="I", einsum="Q")]
        self.assertAlmostEqual(stats.net_total_read_actions(), 64.0)
        self.assertAlmostEqual(stats.net_max_per_unit_read_actions(), 16.0)
        self.assertAlmostEqual(stats.net_total_write_actions(), 64.0)
        self.assertAlmostEqual(stats.net_max_per_unit_write_actions(), 16.0)

        stats = result.buffet_stats[Buffet(level="MainMemory", tensor="Q", einsum="Q")]
        self.assertAlmostEqual(stats.net_total_read_actions(), 0)
        self.assertAlmostEqual(stats.net_max_per_unit_read_actions(), 0)
        self.assertAlmostEqual(stats.net_total_write_actions(), 16.0)
        self.assertAlmostEqual(stats.net_max_per_unit_write_actions(), 16.0)

        stats = result.buffet_stats[Buffet(level="LocalBuffer", tensor="Q", einsum="Q")]
        self.assertAlmostEqual(stats.net_total_read_actions(), 64.0)
        self.assertAlmostEqual(stats.net_max_per_unit_read_actions(), 16.0)
        self.assertAlmostEqual(stats.net_total_write_actions(), 64.0)
        self.assertAlmostEqual(stats.net_max_per_unit_write_actions(), 16.0)


class TestSymbolicActions(unittest.TestCase):
    def test_q_mapping(self):
        mapping = Mapping.from_yaml(Path(__file__).parent / "Q_mapping.yaml")
        workload = Workload.from_yaml(Path(__file__).parent / "mha.yaml")

        result = analyze_reuse_and_add_reservations_to_mapping(
            make_job(mapping, workload)
        )
        actions = gather_actions(result, None, use_name=True)

        self.assertAlmostEqual(actions[("LocalBuffer", "read")].total, 128.0)
        self.assertAlmostEqual(actions[("LocalBuffer", "read")].max_per_unit, 32.0)
        self.assertAlmostEqual(actions[("LocalBuffer", "write")].total, 128.0)
        self.assertAlmostEqual(actions[("LocalBuffer", "write")].max_per_unit, 32.0)

        self.assertAlmostEqual(actions[("MAC", "compute")].total, 64.0)
        self.assertAlmostEqual(actions[("MAC", "compute")].max_per_unit, 16.0)


class TestSymbolicLatency(unittest.TestCase):
    def test_q_mapping(self):
        spec = Spec.from_yaml(
            [
                # Path(__file__).parent / 'Q_mapping.yaml',
                Path(__file__).parent / "mha.yaml",
                Path(__file__).parent / "four_level.arch.yaml",
            ]
        )
        workload = spec.workload
        architecture = spec._get_flattened_architecture()
        mapping = Mapping.from_yaml(Path(__file__).parent / "Q_mapping.yaml")

        result = analyze_reuse_and_add_reservations_to_mapping(
            make_job(mapping, workload)
        )
        overall_latency, _, _ = get_latency(result, mapping, workload, architecture)

        self.assertAlmostEqual(overall_latency, 16.0)


if __name__ == "__main__":
    unittest.main()



"""
Handles the ISL spatial reuse functions.
"""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional

import islpy as isl

from accelforge.frontend.mapping import MappingNode
from accelforge.model._looptree.reuse.isl.isl_functions import (
    insert_equal_dims_map,
    reorder_projector,
)
from accelforge.model._looptree.reuse.isl.mapping_to_isl.types import (
    TEMPORAL_TAGS,
    Fill,
    Occupancy,
    SpatialTag,
    Tag,
    TaggedMap,
)


class Transfers(TaggedMap):
    """Transfers between regions in spacetime."""


class Reads(TaggedMap):
    """Reads between regions in spacetime."""


@dataclass(frozen=True, slots=True)
class TransferInfo:
    """Data transfer information about a certain [subset] of the chip."""

    # Crucial information to transfer info.
    fulfilled_fill: Transfers
    """Fills done by peer-to-peer transfers."""
    unfulfilled_fill: Fill
    """Fills not performed."""
    parent_reads: Reads
    """Fills done by parent-to-child transfers."""
    hops: isl.PwQPolynomial
    """Peer-to-peer transfer cost metric across spacetime."""

    # Metadata on what is occurring.
    link_transfer: bool


class TransferModel(ABC):
    """
    A peer-to-peer/multicast transfer model for spatial analysis.
    """

    @abstractmethod
    def apply(self, buff: MappingNode, fills: Fill, occs: Occupancy) -> TransferInfo:
        """
        Given a buffer, its fills across time, and its occupancies across time,
        calculate the spatial transfers.

        Parameters
        ----------
        buff:
            The buffer whose spatial analysis is being considered.
        fills:
            The fill of `buffer` across time from parents.
        occs:
            The occupancy of `buffer` across time.

        Returns
        -------
        Fills that were fulfilled, Fills that were unfilled, and parent reads per
        position in spacetime. Then, gets hops per timestep.
        """
        raise NotImplementedError(
            f"{type(self)} has not implemented `apply(self, MappingNode, Fill, Occupancy)`"
        )

    def __repr__(self):
        """Returns what transfer model it is."""
        return f"{type(self)}"


class SimpleLinkTransferModel(TransferModel):
    """
    Basic link transfer model.
    """

    def apply(self, buff: MappingNode, fills: Fill, occs: Occupancy) -> TransferInfo:
        # Sanity check the fill is for the same occupancy. Necessary but insufficient proof.
        assert fills.tags == occs.tags, (
            "Fill and Occupancy mismatch"
            "---------------------------"
            f"Fill: {fills}\n"
            f"Occs: {occs}\n"
        )

        # Gets number of input dimensions, along with spatial and temporal indices.
        n: int = fills.map_.dim(isl.dim_type.in_)
        spatial_dims: list[int] = get_spatial_tags_idxs(fills.tags, buff)
        last_temporal: Optional[int] = get_last_temporal_tag_idx(fills.tags)

        # No temporal or no spatial dims, you're just not moving data across time
        # so no transfers occurring.
        if last_temporal is None or len(spatial_dims) == 0:
            return TransferInfo(
                fulfilled_fill=Transfers(
                    fills.tags, fills.map_.subtract(fills.map_)
                ),  # Empty map
                unfulfilled_fill=fills,  # No fulfilled_fills, so only unfulfilled_fills
                parent_reads=Reads(
                    occs.tags, occs.map_.subtract(occs.map_)
                ),  # Empty map
                hops=isl.PwQPolynomial.from_qpolynomial(
                    isl.QPolynomial.zero_on_domain(fills.map_.domain().get_space())
                ),
                link_transfer=True,
            )
        # Gets the connectivity between points in space.
        connectivity: isl.Map = make_mesh_connectivity(
            len(spatial_dims), occs.map_.get_tuple_name(isl.dim_type.in_)
        )
        padded_connectivity: isl.Map = insert_equal_dims_map(
            connectivity, 0, 0, n - len(spatial_dims) - 1
        )
        permutation: list[int] = make_connectivity_permutation(spatial_dims, n)
        reorder_map: isl.Map = reorder_projector(
            permutation, occs.map_.get_tuple_name(isl.dim_type.in_)
        )
        complete_connectivity: isl.Map = reorder_map.apply_range(
            padded_connectivity
        ).apply_range(reorder_map.reverse())

        # Gets data available from neighbors at each point in space per time.
        available_from_neighbors: isl.Map = complete_connectivity.apply_range(occs.map_)
        # Prunes data that does not need to be fetched from a higher in the mem hierarchy.
        neighbor_filled: isl.Map = fills.map_.intersect(available_from_neighbors)

        return TransferInfo(
            fulfilled_fill=Transfers(fills.tags, neighbor_filled.coalesce()),
            unfulfilled_fill=Fill(
                fills.tags, fills.map_.subtract(neighbor_filled).coalesce()
            ),
            # Empty, since only p2p analyzed.
            parent_reads=Reads(
                fills.tags, neighbor_filled.subtract(neighbor_filled).coalesce()
            ),
            hops=isl.PwQPolynomial.from_qpolynomial(
                isl.QPolynomial.one_on_domain(neighbor_filled.wrap().get_space())
            )
            .intersect_domain(neighbor_filled.wrap())
            .coalesce(),
            link_transfer=True,
        )


def make_mesh_connectivity(n: int, spacetime: str) -> isl.Map:
    """
    Makes a neighbor-to-neighbor mesh connection given a number of spatial dims.

    Parameters
    ----------
    n:
        The number of spatial dimensions.
    spacetime:
        The name of the spacetime the mesh is operating on.

    Returns
    -------
    A direct orthogonal adjacency map on the space `spacetime[t, x_1, x_2, ..., x_n]`
    """
    mesh: isl.Map
    match (n):
        case 2:
            mesh = isl.Map.read_from_str(
                isl.DEFAULT_CONTEXT,
                "{ [t, x, y] -> [t-1, x', y'] : "
                " (y'=y and x'=x-1) or (y'=y and x'=x+1) "
                " or (x'=x and y'=y-1) or (x'=x and y'=y+1) }",
            )
        case 1:
            mesh = isl.Map.read_from_str(
                isl.DEFAULT_CONTEXT,
                "{ [t, x] -> [t-1, x'] : (x'=x-1) or (x'=x+1) }",
            )
        case _:
            raise ValueError(f"Cannot make mesh with {n} spatial dims")

    mesh = mesh.set_tuple_name(isl.dim_type.in_, spacetime).set_tuple_name(
        isl.dim_type.out, spacetime
    )

    return mesh


def make_connectivity_permutation(spatial_idxs: list[int], dims: int) -> list[int]:
    """TODO: Figure out what this is doing."""
    permutation: list[int] = []

    cur_spatial_idx: int = 0
    for i in range(dims):
        if cur_spatial_idx < len(spatial_idxs) and i == spatial_idxs[cur_spatial_idx]:
            cur_spatial_idx += 1
        else:
            permutation.append(i)

    for spatial_idx in spatial_idxs:
        permutation.append(spatial_idx)

    return permutation


def get_spatial_tags_idxs(tags: list[Tag], buffer: MappingNode) -> list[int]:
    """
    Given a list if tags, identify the spatial dimensions belong to a given `buffer`.

    Parameters
    ----------
    tags:
        The `Occupancy` or `Fill` domain dimension tags.
    buffer:
        The `MappingNode` which is the logical-memory we're looking for spatial
        dims over.

    Returns
    -------
    A list of the spatial_dim_idxs in order.
    """
    spatial_dim_idxs: list[int] = [
        i
        for i, tag in enumerate(tags)
        if isinstance(tag, SpatialTag) and tag.buffer == buffer
    ]

    return spatial_dim_idxs


def get_last_temporal_tag_idx(tags: list[Tag]) -> Optional[int]:
    """
    Returns the idx of the deepest temporal tag in the list.

    Parameters
    ----------
    tags:
        A list of `Tags`.

    Returns
    -------
    The index of the last tag that is a `TEMPORAL_TAGS`.
    """
    if len(tags) == 0:
        return None

    for idx, tag in reversed(list(enumerate(tags))):
        if isinstance(tag, TEMPORAL_TAGS):
            return idx

    return None



from dataclasses import dataclass
from pathlib import Path


@dataclass
class LoopTreeStatistics:
    latency: float
    energy: float
    actions: dict
    memory_latency: dict
    capacity_usage: dict


def run_symbolic_model(mapping, workload, architecture):
    from pytimeloop._looptree.reuse import analyze_reuse_and_add_reservations_to_mapping
    from pytimeloop._looptree.energy import gather_actions

    job = Job.make_job(mapping=mapping, workload=workload, architecture=architecture)
    result = analyze_reuse_and_add_reservations_to_mapping(job)
    actions = gather_actions(result, bindings, use_name=True)
    pass


def run_looptree(config_dir, paths, tmp_path, bindings, call_accelergy):
    import islpy as isl
    from bindings.config import Config
    from bindings._looptree import LooptreeModelApp, LooptreeWorkload
    from pytimeloop.file import gather_yaml_configs
    from pytimeloop._looptree.capacity import compute_capacity_usage
    from pytimeloop._looptree.reuse._isl.des import deserialize_looptree_output
    from pytimeloop._looptree.energy import gather_actions, compute_energy_from_actions
    from pytimeloop._looptree.latency import get_latency
    from pytimeloop.timeloopfe.v4fused import Spec
    from pytimeloop.timeloopfe.common.backend_calls import call_accelergy_verbose

    yaml_str = gather_yaml_configs(config_dir, paths)
    config = Config(yaml_str, "yaml")
    model = LooptreeModelApp(config)

    workload = LooptreeWorkload.parse_cfg(config.root["problem"])

    spec = Spec.from_yaml_files([str(config_dir / p) for p in paths])

    if call_accelergy:
        if isinstance(tmp_path, Path):
            tmp_path = str(tmp_path)
        call_accelergy_verbose(spec, tmp_path)
        spec = Spec.from_yaml_files(
            [str(config_dir / p) for p in paths] + [str(Path(tmp_path) / "ERT.yaml")]
        )

    result = deserialize_looptree_output(model.run(), isl.DEFAULT_CONTEXT)

    actions = gather_actions(result, bindings)
    energy = compute_energy_from_actions(actions, spec.ERT)

    latency, comp_latency, mem_latency = get_latency(
        result, spec.mapping, workload, spec.arch, bindings
    )

    capacity_usage = compute_capacity_usage(
        spec.mapping.nodes, result.occupancy, workload
    )
    component_capacity_usage = {}
    for level, component in bindings.items():
        if level in capacity_usage:
            component_capacity_usage[component] = capacity_usage[level]

    return LoopTreeStatistics(
        latency, energy, actions, mem_latency, capacity_usage=component_capacity_usage
    )


def run_looptree_symbolic(config_dir, paths, tmp_path, bindings, call_accelergy):
    from bindings.config import Config
    from bindings._looptree import LooptreeWorkload, LooptreeWorkloadDependencyAnalyzer
    from pytimeloop.file import gather_yaml_configs
    from pytimeloop._looptree.capacity import compute_capacity_usage
    from pytimeloop._looptree.reuse import analyze_reuse_and_add_reservations_to_mapping
    from pytimeloop._looptree.energy import gather_actions, compute_energy_from_actions
    from pytimeloop._looptree.latency import get_latency
    from pytimeloop.timeloopfe.v4fused import Spec
    from pytimeloop.timeloopfe.common.backend_calls import call_accelergy_verbose
    from accelforge.mapper.FFM._make_pmappings.pmapper_job import Job

    yaml_str = gather_yaml_configs(config_dir, paths)

    config = Config(yaml_str, "yaml")
    workload = LooptreeWorkload.parse_cfg(config.root["problem"])
    analyzer = LooptreeWorkloadDependencyAnalyzer(workload)

    spec = Spec.from_yaml_files([str(config_dir / p) for p in paths])

    if call_accelergy:
        if isinstance(tmp_path, Path):
            tmp_path = str(tmp_path)
        call_accelergy_verbose(spec, tmp_path)
        spec = Spec.from_yaml_files(
            [str(config_dir / p) for p in paths] + [str(Path(tmp_path) / "ERT.yaml")]
        )

    job = Job.make_job(mapping=spec.mapping, workload=workload, architecture=spec.arch)
    tile_shapes, result = analyze_reuse_and_add_reservations_to_mapping(job)

    actions = gather_actions(result, bindings, use_name=True)
    energy = compute_energy_from_actions(actions, spec.ERT)

    latency, comp_latency, mem_latency = get_latency(
        result, spec.mapping, workload, spec.arch, bindings
    )

    capacity_usage = compute_capacity_usage(
        spec.mapping.nodes, result.occupancy, workload
    )
    component_capacity_usage = {}
    for level, component in bindings.items():
        if level in capacity_usage:
            component_capacity_usage[component] = capacity_usage[level]

    return LoopTreeStatistics(
        latency, energy, actions, mem_latency, capacity_usage=component_capacity_usage
    )



"""
Models for handling calculating the cost of a Workload on distributed buffer
architectures.
"""

import logging

import islpy as isl

from accelforge.frontend.mapping import MappingNode
from accelforge.model._looptree.reuse.isl.isl_functions import dim_projector_mask
from accelforge.model._looptree.reuse.isl.mapping_to_isl import DUMP_ISL_IR
from accelforge.model._looptree.reuse.isl.mapping_to_isl.types import Fill, Occupancy
from accelforge.model._looptree.reuse.isl.spatial import (
    Reads,
    Transfers,
    TransferInfo,
    TransferModel,
)


def identify_mesh_casts(
    src_occupancy: isl.Map, dst_fill: isl.Map, dist_fn: isl.Map
) -> isl.Map:
    """
    Given srcs with data, fills to destinations, and a distance function, identify per data
    the srcs delivering that data to dsts.

    Parameters
    ----------
    src_occupancy:
        An isl.Map of the form { [src] -> [data] } corresponding to the data held
        at the buffer at space `src`.
    dst_fill:
        An isl.Map of the form { [dst] -> [data] } corresponding to the data requested
        at the element at space `dst`.
    dist_fn:
        A distance function { [src -> dst] -> [hops] } that accepts two points in
        space, corresponding to the `src` and `dst`, and returns the distance
        between the two points in terms of `hops`, a quantized atomic distance of
        data transmission cost.

    Returns
    -------
    { [data] -> [dst -> src] } where { [dst] -> [data] } and { [src] -> [data] } are in
    `src_occupancy` and `dst_fill` respectively, and where `[dst -> src]` is the infimum of
    `dst_fn(src, dst), ∀ src, dst s.t. { [src] -> [data] } ∈ `src_occupancy` and
    `{ [dst] -> [data] }` ∈ `dst_fill`.
    """
    # Makes { [dst -> data] -> [dst -> data] }
    fill_to_fill: isl.Map = dst_fill.wrap().identity()
    if DUMP_ISL_IR:
        logging.info(f"fill_to_fill: {fill_to_fill}")

    # Inverts src_occupancy s.t. data -> src.
    # i.e. { [xs, ys] -> [d0, d1] } to { [d0, d1] -> [xs, ys] }
    data_presence: isl.Map = src_occupancy.reverse()

    # { [dst -> data] -> [dst -> src] } where src contains data.
    fills_to_matches: isl.Map = (
        fill_to_fill.uncurry()  # { [[dst -> data] -> dst] -> data }
        .apply_range(data_presence)  # { [[dst -> data] -> dst] -> src }
        .curry()
    )  # { [[dst -> data] -> [dst -> src] }
    if DUMP_ISL_IR:
        logging.info(f"fills_to_matches: {fills_to_matches}")

    # Calculates the distance of a fill to the nearest src satisfying the fill.
    # { [dst -> data] -> [dist] }
    fill_min_dist: isl.Map = fills_to_matches.apply_range(dist_fn).lexmin()
    # Isolates the relevant minimal pairs.
    # { [dst -> data] -> [dst -> src] :.dst -> src is minimized distance }
    minimal_pairs: isl.Map = (
        fill_min_dist.apply_range(
            # Note: Need to match fill -> min_dist with min_dist -> [fill -> match] as lexmin over
            # fill and match will minimize distance over the tuple (src, dst, data), but that
            # overconstrains the optimization as we want to minimize over distance (dst, data)
            # only for all src.
            fills_to_matches.range_map()
            .apply_range(dist_fn)
            .reverse()
        )
        .range()
        .unwrap()
    )
    if DUMP_ISL_IR:
        logging.info(f"minimal_pairs: {minimal_pairs}")

    # Isolates the multicast networks.
    # { [data] -> [dst -> src] : dst -> src is minimized distance }
    multicast_networks: isl.Map = minimal_pairs.curry().range().unwrap()
    # Devolves to a single source if multiple sources per domain point.
    multicast_networks = multicast_networks.uncurry().lexmin().curry()

    return multicast_networks


def calculate_extents_per_dim(mcns: isl.Map) -> list[isl.PwAff]:
    """
    Parameters
    ----------
    mcns:
        Mesh cast-networks, or networks in which all dsts per data are grouped with
        the closest src containing the data.

    Returns
    -------
    A list of `isl.PwAff` that gives the max extent (length) along dim_i per mcn,
    where i is the i-th `isl.PwAff`.

    Preconditions
    -------------
    `mcns` were generated with a Manhattan distance `dst_fn` by `identify_mesh_casts`
    s.t. all dimensions are orthogonal to each other in a metric space, where each
    unit movement in a dimension counts as 1 hop.

    We also assume `dst_fn` is translationally invariant (i.e., ∀src, dst,
    src', dst' ∈ space, if |src - dst| = |src' - dst'|,
    dst_fn(src, dst) = dst_fn(src', dst').
    """
    # Makes mcns from { [data] -> [dst -> src] } to { [data -> src] -> [dst] }
    potential_srcs: isl.Map = mcns.range_reverse().uncurry()
    # Sources are part of the extents, so we union it with the destinations.
    # { [data -> src] -> [src] }
    srcs: isl.Map = potential_srcs.domain().unwrap().range_map()
    # { [data -> src] -> [spacetime] }
    casting_extents: isl.Map = srcs.union(potential_srcs)

    # Projects away all dimensions but one to find their extent for hypercube.
    dims: int = potential_srcs.range_tuple_dim()
    # Creates a mask of what to project out.
    project_out_mask: list[bool] = [True] * dims
    dim_extents: list[isl.PwAff] = [None] * dims

    # Gets the extents of all dimensions
    for noc_dim in range(dims):
        # Project out all the dimensions of the output besides noc_dim.
        project_out_mask[noc_dim] = False
        # { [spacetime] -> [dimension] }
        extent_mapper: isl.Map = dim_projector_mask(
            casting_extents.range().get_space(), project_out_mask
        ).reverse()
        dim_extent_space: isl.Map = casting_extents.apply_range(extent_mapper)
        project_out_mask[noc_dim] = True

        # Finds max(noc_dim) - min(noc_dim) for each [data -> src]
        max_extent: isl.PwAff = dim_extent_space.dim_max(0)
        min_extent: isl.PwAff = dim_extent_space.dim_min(0)

        # Subtracts the max from the min to get the extent per [data -> src]
        dim_extents[noc_dim] = max_extent.sub(min_extent).coalesce()

    return dim_extents


class HypercubeMulticastModel(TransferModel):
    """
    Does distributed multicasting a mesh using worst-case multicasting
    behavior by assuming all multicasts are broadcasting to the convex
    hypercube that encapsulates all their destinations and sources.
    """

    def __init__(self, dist_fn: isl.Map):
        """
        Initializes the HypercubeMulticastModel with the distance function
        over the metric space.

        Because we are using calculate_extents_per_dim(mcns), we inherit the
        following requirements:
        `dst_fn` holds all dimensions are orthogonal to each other in a metric space,
        where each unit movement in a dimension counts as 1 hop.

        We also assume `dst_fn` is translationally invariant (i.e., ∀src, dst,
        src', dst' ∈ space, if |src - dst| = |src' - dst'|,
        dst_fn(src, dst) = dst_fn(src', dst').
        """
        self.dist_fn = dist_fn

    def apply(self, buff: MappingNode, fills: Fill, occs: Occupancy) -> TransferInfo:
        """
        Given a buffer, its fills across time, and its occupancies across time,
        calculate the spatial transfers."

        Parameters
        ----------
        buff:
            The buffer whose spatial analysis is being considered. Currently,
            we rely on dist_fn to deal with this rather than buffer.
        fills:
            The fill of `buffer` across time from parents.
        occs:
            The occupancy of `buffer` across time.

        Returns
        -------
        Fills that were fulfilled, Fills that were unfilled, and parent reads per
        position in spacetime. Then, gets hops per timestep.
        """
        mcs: isl.Map = identify_mesh_casts(occs.map_, fills.map_, self.dist_fn)
        result: isl.PwQPolynomial = self._cost_mesh_cast_hypercube(mcs)

        # TODO: Read once from all buffers, assert that
        # card(mcs) == tensor_size * duplication factor
        num_meshcasts: int = mcs.card()
        return TransferInfo(
            fulfilled_fill=Transfers(fills.tags, fills.map_),
            parent_reads=Reads(occs.tags, mcs),
            unfulfilled_fill=Fill(fills.tags, fills.map_.subtract(fills.map_)),
            hops=result,
            link_transfer=True,
        )

    def _cost_mesh_cast_hypercube(self, mcns: isl.Map) -> int:
        """
        Given a multicast_network, calculate the hypercube.

        Parameters
        ----------
        mcns:
            Multicast networks grouped together by [srcs -> data] fulfilling
            [dsts -> data], where there is at least 1 src and 1 dst in each mcn.

        Returns
        -------
        The upperbound of doing all the multicasts specified by the multicast
        networks, assuming they cast only to the convex space of the network.

        Preconditions
        -------------
        Because we are using calculate_extents_per_dim(mcns), we inherit the
        following requirements:
        `mcns` were generated with a Manhattan distance `dst_fn` by `identify_mesh_casts`
        s.t. all dimensions are orthogonal to each other in a metric space, where each
        unit movement in a dimension counts as 1 hop.

        We also assume `dst_fn` is translationally invariant (i.e., ∀src, dst,
        src', dst' ∈ space, if |src - dst| = |src' - dst'|,
        dst_fn(src, dst) = dst_fn(src', dst').
        """
        dim_extents: list[isl.PwAff] = calculate_extents_per_dim(mcns)
        # Tracks the total cost of the hypercube cast per [src -> data]
        one: isl.PwAff = isl.PwAff.val_on_domain(dim_extents[0].domain(), 1)
        hypercube_costs = isl.PwQPolynomial.from_pw_aff(one)

        # Calculates the cost of the hypercube, where the hypercube cost
        # = \sum_{i=0}^{D} ((extent_i - 1) * \prod_{j=0}^{i-1} extent_j)
        # = (\prod_{i=0}^{D} extent_i) - 1
        for dim_extent in dim_extents:
            # Adds the dim_extent times the casting volume to the hypercube
            # cost.
            dim_plus: isl.PwQPolynomial = isl.PwQPolynomial.from_pw_aff(
                dim_extent.add(one).coalesce()
            )
            hypercube_costs = hypercube_costs.mul(dim_plus).coalesce()
        hypercube_costs = hypercube_costs.sub(isl.PwQPolynomial.from_pw_aff(one))

        # Tracks the total cost of the hyppercube cast per data.
        hypercube_costs = hypercube_costs.sum()

        # Return the hypercube cost as a piecewise polynomial.
        return hypercube_costs.sum()



"""
Utility functions common to testing the isl mapper functions.
"""

from pathlib import Path
import islpy as isl

from ruamel.yaml import YAML


def to_isl_maps(obj: str | list | dict) -> dict:
    """
    Given an object, attempt to reduce all strings in tree with isl.Map

    Parameters
    ----------
    obj:
        A DAG which can be explored and contains isl.Map strings within it.

    Returns
    -------
    `obj` but all strings are converted to isl.Map.
    """

    def _to_isl_maps(obj: str | dict | list) -> isl.Map | dict | list:
        """Recursively convert string ISL maps to isl.Map; leave others alone."""
        if isinstance(obj, str):
            return isl.Map.read_from_str(isl.DEFAULT_CONTEXT, obj)
        if isinstance(obj, dict):
            return {k: (_to_isl_maps(v) if k != "type" else v) for k, v in obj.items()}
        if isinstance(obj, list):
            return [_to_isl_maps(v) for v in obj]
        return obj

    return _to_isl_maps(obj)  # type: ignore


def load_solutions(path: Path) -> dict:
    """
    Loads in a dictionary with the isl solutions to a workload problem.

    Parameters
    ----------
    path:
        The path to the solutions.

    Returns
    -------
    A dictionary relating Python-based keys generated by the mapper (e.g.,
    `BufferTensorEinsum` to their corresponding isl.Map.)
    """
    # Load expected solutions (YAML file with string ISL maps)
    yaml: YAML = YAML(typ="safe")

    with open(path, "r", encoding="utf-8") as f:
        return to_isl_maps(yaml.load(f))



from accelforge.util._basetypes import EvalableModel
from accelforge.frontend.mapper.ffm import FFM


class Mapper(EvalableModel):
    ffm: FFM = FFM()
    """ Fast and Fusiest Mapper configuration. Currently the only supported mapper. """
